import matplotlib.pyplot as plt
import numpy
import keysightSD1
import keysight_hvi as kthvi
import keysight_ktm5302x as denim

"""
Pathwave Test Sync Executive

Copyright Keysight Technologies 2022

In this programming example, Pathwave Sync Executive is used to program multiple M3xxxA AWGs and M5302A Digital I/O (DIO) instruments.
By using HVI (Hard Virtual Instrument) capabilities, DIO instruments can output a pulsed signal from SMB trigger ports and AWGs can 
synchronously play a previously queued waveform. All modules run fully synchronized and actions across modules can be controlled with 
the timing resolution of the M3xxxA AWGs which is 10ns.

The showcased HVI API 2021 functionalities are:
* Scalability to an arbitrary number of instruments. In this specific case M3xxxA PXI AWGs and M5302A Digital I/O instruments.
* Off-shelf inter-instrument synchronization capabilities
* HVI Synchronized Multi-Sequence Block to synchronously parallel-execute sequences in multiple instruments
* HVI core and HVI product-specific instructions

The Python code that follows: 
1. Opens an arbitrary number of M3xxxA AWGs (real hardware or simulation mode) 
2. Defines HVI resources (engines, chassis, chassis interconnections, trigger, action, events, HVI triggering resources, etc)
3. Programs HVI sequences by adding the necessary statements and instructions in each AWG's sequence
4. Compiles HVI, loads it to HW and executes it

For further info please check the programming example titled "Synchronized MIMO Measurements using M5302A Digital I-O and M3xxxA PXI Instruments"
"""

######################################
# Application Configuration 
######################################

class ApplicationConfig:
    """
    Defines module descriptors, configuration options and names of HVI engines, actions, triggers
    """
    def __init__(self):
        """
        Define PXI instruments options
        """
        # Simulated HW Option
        # Set hardware_simulated to True to run the example in simulation mode
        self.hardware_simulated = True
        # Define options to open the M3xxxA instruments
        # Complete option list can be found in the SD1 3.x User Guide
        self.sd1_options = 'channelNumbering=keysight'
        # Define options to open the M5302A Digital I/O instruments
        self.dio_query = True
        self.dio_reset = True
        self.dio_options = 'Simulate=0, LogLevel=Info'
        self.dio_model_number = "M5302A"
        """ 
        Define names of HVI engines, actions, triggers, registers
        """
        # The leader engine is assigned to be the first one in the dio_module_descriptors list
        self.leader_engine_name = ""
        self.awg_engine_name = "AwgEngine"
        self.dig_engine_name = "DigEngine"
        self.dio_engine_name = "DioEngine"
        self.awg_trigger_action_name = "AwgTrigger"
        self.dig_trigger_action_name = "DaqTrigger"
        self.smb_trigger_name = "SmbTrigger"
        self.loop_register_name = "Loops"
        self.num_loops = 5
        # Note: the digitizer has a trigger re-arming time greater than 1.8 us
        # Loop duration cannot be smaller than the digitizer re-arming time
        # Please check the SD1 3.x Digitizer User Manual for exact latency values
        self.loop_duration = 2000 # [ns]
        """
        Define HW Platform
        """
        # Define module descriptors below with your instruments information
        self.dig_module_descriptors = [
            Sd1ModuleDescriptor('M3102A', 1, 15, self.sd1_options, self.dig_engine_name)]
        self.awg_module_descriptors = [
            Sd1ModuleDescriptor('M3202A', 1, 5, self.sd1_options, self.awg_engine_name)]
        self.dio_module_descriptors = [
            DioModuleDescriptor('PXI0::CHASSIS1::SLOT16::INDEX0::INSTR', self.dio_query, self.dio_reset, self.dio_options, self.dio_model_number, self.dio_engine_name),
            DioModuleDescriptor('PXI0::CHASSIS1::SLOT17::INDEX0::INSTR', self.dio_query, self.dio_reset, self.dio_options, self.dio_model_number, self.dio_engine_name)]
        # Combines all previous descriptors
        self.module_descriptors = []
        self.define_module_descriptors()
        # Defines leader engine and numbers the other engines
        self.define_engine_names()
        
        # Define list of chassis containing instruments to be used by HVI
        self.chassis_list = [1]
        
        # Multi-chassis setup
        # Define the System Sync Modules included in your system.
        # For example: 
        # self.system_sync_module_descriptors = [
        #    SystemSyncModuleDescriptor('PXI0::CHASSIS2::SLOT10::INDEX0::INSTR', self.ssm_options)]
        self.ssm_options = ''
        self.ssm_simulation_options = 'Simulate=true,DriverSetup=Model=M9033A'
        self.system_sync_module_descriptors = []
        # For each SSM define which SSM is connected to its downstream connectors.
        # For example: 
        # self.ssm_connections = [
        #      SystemSyncModuleConnection(ssm1_chassis=1, ssm1_downstream_connector_number=3, ssm2_chassis=2)]
        self.ssm_connections = []
        
        # Assign triggers to HVI object to be used for HVI-managed synchronization, data sharing, etc
        # NOTE: In a multi-chassis setup ALL the PXI lines listed below need to be shared among each M9031 board pair by means of SMB cable connections
        self.pxi_sync_trigger_resources = [
            kthvi.TriggerResourceId.PXI_TRIGGER5, 
            kthvi.TriggerResourceId.PXI_TRIGGER6, 
            kthvi.TriggerResourceId.PXI_TRIGGER7]
        """
        AWG Parameters
        """
        # AWG settings for all channels
        self.sync_mode = keysightSD1.SD_SyncModes.SYNC_NONE
        self.queue_mode = keysightSD1.SD_QueueMode.ONE_SHOT
        self.awg_start_delay = 0
        self.awg_prescaler = 0
        self.awg_num_cycles = 0
        self.amplitude = 1
        # Trigger settings
        self.awg_trigger_mode = keysightSD1.SD_TriggerModes.SWHVITRIG_CYCLE
        """
        Digitizer parameters
        """
        self.dig_sampling_time = 2 # [ns] 1/sample_rate, sample_rate = 500 MSa/s for Digitizer M3102A 
        self.acquisition_points_per_cycle = 300
        self.dig_prescaler = 0
        self.fullscale = 2 # [V] enter x Volts to set the full scale to [-x, x] Volts
        self.acquisition_time_per_cycle = self.acquisition_points_per_cycle*self.dig_sampling_time
        self.dig_num_cycles = self.num_loops #insert -1 for infinite cycles
        self.acquisition_points = int(self.acquisition_points_per_cycle*self.num_loops)
        self.acquisition_delay = 0 # x2 [ns]
        self.dig_trigger_mode = keysightSD1.SD_TriggerModes.SWHVITRIG
        self.mask = 0xF
    
    def define_module_descriptors(self):
        """
        Combines all modules descriptor into a single module_descriptors array
        """
        # Compose array of module descriptors
        for dig in self.dig_module_descriptors:
            self.module_descriptors.append(dig)
        for awg in self.awg_module_descriptors:
            self.module_descriptors.append(awg)
        for dio in self.dio_module_descriptors:
            self.module_descriptors.append(dio)
    
    def define_engine_names(self):
        """
        Define names for all HVI engines
        """
        # module descriptor lists previously populated in application config class
        for index in range(len(self.awg_module_descriptors)): 
            self.awg_module_descriptors[index].engine_name += str(index)
        for index in range(len(self.dig_module_descriptors)): 
            self.dig_module_descriptors[index].engine_name += str(index)
        for index in range(len(self.dio_module_descriptors)): 
            self.dio_module_descriptors[index].engine_name += str(index)
        # Choose leader engine
        self.leader_engine_name = self.dio_module_descriptors[0].engine_name


######################################
# Main Program
######################################

def main():
    ###################################################
    # Open PXI modules, setup AWGs and queue waveforms
    ###################################################
    # Load configuration
    config = ApplicationConfig()

    # Checks flags defining if the code executes on HW or Simulation mode
    if config.hardware_simulated:
        print("Code running in Simulation Mode")

    # Open instruments
    awg_module_dict = open_awg_modules(config)
    dig_module_dict = open_dig_modules(config)
    dio_module_dict = open_dio_modules(config)
    module_dict = {}
    for engine_name, module in awg_module_dict.items():
        module_dict[engine_name] = module
    for engine_name, module in dig_module_dict.items():
        module_dict[engine_name] = module
    for engine_name, module in dio_module_dict.items():
        module_dict[engine_name] = module
    
    # Configure each AWG and queue waveforms
    for module in awg_module_dict.values():
        configure_awg(config, module.instrument, module.num_channels)

    # Configure each digitizer
    for module in dig_module_dict.values():
        configure_digitizer(config, module.instrument, module.num_channels)

    ######################################
    # System definition
    ######################################
    # Create system definition object
    my_system = kthvi.SystemDefinition("MySystem") 
    # Define your system, HW platform, add HVI resources
    define_hvi_resources(config, my_system, awg_module_dict, dig_module_dict, dio_module_dict)
    # Output to console the details of the system definition
    output_sys_def(my_system, config.module_descriptors)
    
    ######################################
    # Program HVI sequences
    ######################################
    print("-----------------------")
    print("Program HVI Sequences")
    print("-----------------------")
    print("Initializing the defined system...")
    # Create sequencer object
    sequencer = kthvi.Sequencer("MySequencer", my_system)
    print("Programming the HVI sequences...")
    # Program HVI sequences using the sequencer object
    program_mimo_meas_sequence(sequencer, config)
    # Export the programmed sequence to text
    export_hvi_sequences(sequencer)
    
    ########################################
    # Compile, Load to HW, Run HVI Sequence
    ########################################
    print("-------------------")
    print("Execute HVI")
    print("-------------------")
    print("Compiling HVI...")
    # Compile HVI sequences
    hvi = sequencer.compile()
    print(hvi.compile_status.to_string())
    print("HVI Compiled")
    print("This HVI needs to reserve {} PXI trigger resources to execute".format(len(hvi.compile_status.sync_resources)))
    
    # Load HVI to HW: load sequences, configure actions/triggers/events, lock resources, etc.
    hvi.load_to_hw()    
    print("HVI Loaded to HW")

    # Execute HVI in blocking mode: SW waits until HVI sequences ends their execution
    # Eventually enter a timeout for the HVI execution to be stopped: timeout = timedelta(seconds=0), hvi.run(timeout)
    hvi.run(hvi.no_timeout)
    print("HVI Running...")

    # User-controlled HVI execution
    if config.hardware_simulated:
        print("Simulation completed successfully")
    else:
        interact_with_hvi(hvi, dig_module_dict[config.dig_engine_name+str(0)].instrument, config)

    # Release HW resources once HVI execution is completed
    hvi.release_hw()
    print("Releasing HW...")

    # Close all modules at the end of the execution
    for engine_name in module_dict:
        module_dict[engine_name].instrument.close()
    print("PXI modules closed\n")

########################################################
# 2nd Level: Functions to Define, Program, Execute HVI
#######################################################

def define_hvi_resources(config, sys_def, awg_module_dict, dig_module_dict, dio_module_dict):
    """
    Configures all the necessary resources for the HVI application to execute: HW platform, engines, actions, triggers, etc. 
    """
    # Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks
    define_hw_platform(sys_def, config)

    # Define all the HVI engines to be included in the HVI
    define_hvi_engines(sys_def, awg_module_dict, dig_module_dict, dio_module_dict)
    
    # Define list of actions to be executed
    define_hvi_actions(sys_def, awg_module_dict, dig_module_dict, config)

    # Defines the trigger resources
    define_hvi_triggers(sys_def, dio_module_dict, config)


def define_hw_platform(sys_def, config):
    """
    Define HW platform: chassis, interconnections, PXI trigger resources, synchronization, HVI clocks 
    """
    # Define chassis resources 
    # For multi-chassis setup details see programming example documentation
    for chassis_number in config.chassis_list:
        if config.hardware_simulated:
            sys_def.chassis.add_with_options(chassis_number, 'Simulate=True,DriverSetup=Model=GenericPxieChassis')
        else:
            sys_def.chassis.add(chassis_number)

    # Define System Sync Modules (SSMs)
    if config.system_sync_module_descriptors:
        interconnects = sys_def.interconnects
        ssm_list = []
        for descriptor in config.system_sync_module_descriptors:
            if config.hardware_simulated:
                ssm = interconnects.add_sync_module(descriptor.resource_id, config.ssm_simulation_options)
            else:
                ssm = interconnects.add_sync_module(descriptor.resource_id, descriptor.options)
            # Append SSM object
            ssm_list.append(ssm)
        # Define connections between SSMs
        if config.ssm_connections:
            for connection in config.ssm_connections:
                connector_number = connection.ssm1_downstream_connector_number
                for ssm in ssm_list:
                    if ssm.chassis == connection.ssm1_chassis:
                        ssm1 = ssm
                    if ssm.chassis == connection.ssm2_chassis:
                        ssm2 = ssm
                # Implement each user-defined connection
                try:
                    # Set connection. SSMs have always one upstream port
                    ssm1.connectivity.systemsync_downstream[connector_number].set_connection(ssm2.connectivity.systemsync_upstream[0])
                except:
                    exit("Exception! Please check the valued defined for SyncModule resource ids, chassis numbers and connections")

    # Assign the defined PXI trigger resources
    sys_def.sync_resources = config.pxi_sync_trigger_resources
    
    # Assign clock frequencies that are outside the set of the clock frequencies of each HVI engine
    # Use the code line below if you want the application to be in sync with the 10 MHz clock
    sys_def.non_hvi_core_clocks = [10e6]


def define_hvi_engines(sys_def, awg_module_dict, dig_module_dict, dio_module_dict):
    """
    Define all the HVI engines to be included in the HVI
    """
    # For each instrument to be used in the HVI application add its HVI Engine to the HVI Engine Collection 
    for engine_name, module in dio_module_dict.items():
        sys_def.engines.add(module.hvi.engines.primary, engine_name)
    for engine_name, module in awg_module_dict.items():
        sys_def.engines.add(module.hvi.engines.main_engine, engine_name)
    for engine_name, module in dig_module_dict.items():
        sys_def.engines.add(module.hvi.engines.main_engine, engine_name)


def define_hvi_actions(sys_def, awg_module_dict, dig_module_dict, config):
    """ 
    Defines AWG/DAQ trigger actions for each module, to be executed by the "action execute" instruction in the HVI sequence
    Create a list of AWG/DAQ trigger actions for each AWG/Digitizer module. The list depends on the number of channels 
    """
    # For each AWG, define the list of HVI Actions to be executed and add such list to its own HVI Action Collection
    for engine_name in awg_module_dict.keys():
        for ch_index in range(1, awg_module_dict[engine_name].num_channels + 1):
            # Actions need to be added to the engine's action list so that they can be executed
            action_name = config.awg_trigger_action_name + str(ch_index) # arbitrary user-defined name
            instrument_action = "awg{}_trigger".format(ch_index) # name decided by instrument API
            action_id = getattr(awg_module_dict[engine_name].hvi.actions, instrument_action)
            sys_def.engines[engine_name].actions.add(action_id, action_name)
    
    # For each Digitizer, define the list of HVI Actions to be executed and add such list to its own HVI Action Collection
    for engine_name in dig_module_dict.keys():
        for ch_index in range(1, dig_module_dict[engine_name].num_channels + 1):
            # Actions need to be added to the engine's action list so that they can be executed
            action_name = config.dig_trigger_action_name + str(ch_index) # arbitrary user-defined name
            instrument_action = "daq{}_trigger".format(ch_index) # name decided by instrument API
            action_id = getattr(dig_module_dict[engine_name].hvi.actions, instrument_action)
            sys_def.engines[engine_name].actions.add(action_id, action_name)


def define_hvi_triggers(sys_def, module_dict, config):
    """
    Defines and configure the eight SMB trigger outputs of each DIO instrument 
    """
    # Add to the HVI Trigger Collection of each DIO HVI Engine the SMB Trigger object of each channel of that same instrument
    for engine_name in module_dict.keys():
        for ch_index in range(1, module_dict[engine_name].num_channels + 1):
            smb_trigger_ch_index = "smb{}".format(ch_index)
            smb_trigger_id = getattr(module_dict[engine_name].hvi.triggers, smb_trigger_ch_index)
            smb_trigger = sys_def.engines[engine_name].triggers.add(smb_trigger_id, config.smb_trigger_name+str(ch_index))
            # Configure SMBx trigger in each hvi.engines[index]
            smb_trigger.config.direction = kthvi.Direction.OUTPUT
            smb_trigger.config.polarity = kthvi.Polarity.ACTIVE_HIGH
            smb_trigger.config.sync_mode = kthvi.SyncMode.IMMEDIATE
            smb_trigger.config.hw_routing_delay = 0
            # NOTE: the trigger mode is set to LEVEL. The length of the trigger pulse is defined by the HVI Instructions TriggerWrite
            smb_trigger.config.trigger_mode = kthvi.TriggerMode.LEVEL


def output_sys_def(sys_def, descriptors):
    """
    Outputs the details of each instrument and engine included in the system definition.
    """
    print("-------------------")
    print("System Definition")
    print("-------------------")
    # Output KS2201A HVI Core Version
    hvi_core_version = kthvi.SystemDefinition.hvi_core_version
    print("The PathWave Test Sync Executive API installed on your system has an HVI core version " + hvi_core_version.to_string() + "\n")
    # Output info about the HW instruments
    print("HW Instruments:")
    for descriptor in descriptors:
        engine = sys_def.engines[descriptor.engine_name]
        print("- Model: {}, HVI Engine Name: {}, HVI core version: ".format(descriptor.model_number, descriptor.engine_name) + 
        engine.software_version.to_string() + ", HVI Engine FPGA IP version: " + engine.firmware_version.to_string())
    # Output info about the Sync Modules
    if sys_def.interconnects:
        print("System Sync Modules:")
        for index in range(sys_def.interconnects.count):
            sync_module = sys_def.interconnects[index]
            print("- System Sync Module in chassis: {}, slot: {}".format(sync_module.chassis, sync_module.slot))


def program_mimo_meas_sequence(sequencer, config):
    """ 
    Programs the MIMO Measurements HVI sync sequence
    """
    # Define loop register
    loops = sequencer.sync_sequence.scopes[config.leader_engine_name].registers.add(config.loop_register_name, kthvi.RegisterSize.SHORT)
    loops.initial_value = 0

    # Define Sync While condition
    condition = kthvi.Condition.register_comparison(loops, kthvi.ComparisonOperator.LESS_THAN, config.num_loops)
    # Add a Sync While
    sync_while = sequencer.sync_sequence.add_sync_while("Sync MIMO Measurements", 90, condition)
    # Set duration of each Sync While iteration
    sync_while.duration = kthvi.time.Duration(config.loop_duration, kthvi.time.Unit.NANOSECONDS)

    # Add a Sync Multi-Sequence Block (SMSB)
    sync_block = sync_while.sync_sequence.add_sync_multi_sequence_block("Trigger Digital I/Os, AWGs and Digitizers", 250)
    # Set SMSB duration
    sync_block.duration = kthvi.time.Minimum()
    # Program the SMSB to trigger AWGs
    program_mimo_trigger(sync_block, config)


def program_mimo_trigger(sync_block, config):
    """ 
    Programs the Sync Multi-Sequence Block to trigger DIOs, AWGs and Digitizers 
    """
    # Add instructions to each local sequence within the multi-sequence block
    for index in range(0, sync_block.sequences.count):
        # Retrieve the engine sequence from the collection of HVI Local Sequences
        # HVI Local Sequence Collection is automatically created form the user-defined HVI Engine Collection
        # Each HVI Local Sequence can be retrieved using the name alias of the corresponding HVI Engine
        sequence = sync_block.sequences[index]
        engine_name = sequence.engine.name
        min_length = min(len(engine_name), len(config.dio_engine_name), len(config.dig_engine_name))
        # Program trigger ON/OFF instructions in the local HVI sequences of DIgital I/O engines
        if engine_name[:min_length] == config.dio_engine_name[:min_length]:
            # Retrieve SMB Triggers prefiously defined in the HVI Trigger Collection
            smb_triggers = sequence.engine.triggers
            # Retrive TriggerWrite instruction from HVI Native InstructionSet
            trigger_write = sequence.instruction_set.trigger_write
            # Create vector of values to be written to each SMB trigger
            on_values = []
            off_values = []
            for trigger in range(0, smb_triggers.count):
                on_values.append(trigger_write.value.on)
                off_values.append(trigger_write.value.off)
            # Write SMB Triggers ON
            instr_trigger_ON = sequence.add_instruction("SMB Triggers ON", 10, trigger_write.id)
            instr_trigger_ON.set_parameter(trigger_write.trigger.id, smb_triggers)
            instr_trigger_ON.set_parameter(trigger_write.sync_mode.id, trigger_write.sync_mode.immediate)
            instr_trigger_ON.set_parameter(trigger_write.value.id, on_values)
            # Write SMB Triggers OFF
            instr_trigger_OFF = sequence.add_instruction("SMB Trigger OFF", 100, trigger_write.id)
            instr_trigger_OFF.set_parameter(trigger_write.trigger.id, smb_triggers)
            instr_trigger_OFF.set_parameter(trigger_write.sync_mode.id, trigger_write.sync_mode.immediate)
            instr_trigger_OFF.set_parameter(trigger_write.value.id, off_values)
        # Program AWG/DAQ Trigger action execution in the local HVI sequences of AWG and Digitizer engines
        else:
            # Define a different instruction label for AWGs or digitizers
            if engine_name[:min_length] == config.dig_engine_name[:min_length]:
                label = "DAQ Trigger"
            else:
                label = "AWG Trigger"
            # "Action Execute" instruction executes the AWG/DAQ trigger from HVI
            action_list = sequence.engine.actions
            instruction = sequence.add_instruction(label, 10, sequence.instruction_set.action_execute.id)
            instruction.set_parameter(sequence.instruction_set.action_execute.action.id, action_list)

        # Add register increment and delay to the leader engine
        if sequence.engine.name == config.leader_engine_name:
            # Retrieve register
            loops = sequence.scope.registers[config.loop_register_name]
            # Increment register
            instruction = sequence.add_instruction("Loops++", 10, sequence.instruction_set.add.id)
            instruction.set_parameter(sequence.instruction_set.add.destination.id, loops)
            instruction.set_parameter(sequence.instruction_set.add.left_operand.id, loops)
            instruction.set_parameter(sequence.instruction_set.add.right_operand.id, 1)
            # Delay statement
            sequence.add_delay("Loop Delay", 50)


def export_hvi_sequences(sequencer):
    """
    Exports the programmed HVI sequences to text format
    """
    # Generate HVI sequence description text
    output = sequencer.sync_sequence.to_string(kthvi.OutputFormat.DEBUG)
    print("Programmed HVI sequences exported to file")
    
    # Write file
    text_file = open("TextOutput/HviSyncMimoSd1Dio.txt", "w")
    text_file.write(output)
    text_file.close()


def interact_with_hvi(hvi, digitizer_module, config):
    """
    User-controlled HVI execution
    If running with hardware, iteratively re-run HVI until the user decides to exit
    """
    # Open figure to display digitizer acquisiions
    plt.ion() #interactive mode
    # fig = plt.figure(1) #open figure
    # plt.show(block=False) #plt.show()
    # fig.canvas.draw()
    while True:
        print("Press enter to run HVI again, q to exit...")
        # DAQstart: digitizer channels get strted and are waiting for the DAQ trigger commands that trigger the acquisition of each cycle of data points
        digitizer_module.DAQstartMultiple(config.mask)
        if input() == 'q':
            # DAQstop
            digitizer_module.DAQstopMultiple(config.mask)
            break
        hvi.run(hvi.no_timeout)
        # Digitizer acquisition
        timeout = 500 #[ms]
        measured_data1 = numpy.zeros(config.acquisition_points)
        measured_data1 = digitizer_module.DAQread(1, config.acquisition_points, timeout)
        measured_data2 = numpy.zeros(config.acquisition_points)
        measured_data2 = digitizer_module.DAQread(2, config.acquisition_points, timeout)
        measured_data3 = numpy.zeros(config.acquisition_points)
        measured_data3 = digitizer_module.DAQread(3, config.acquisition_points, timeout)
        measured_data4 = numpy.zeros(config.acquisition_points)
        measured_data4 = digitizer_module.DAQread(4, config.acquisition_points, timeout)
        # Plot acquisition data
        # Use plt.clf() to clear the plot before each new measurement
        plt.clf()
        plt.plot(measured_data1, 'b-', measured_data2, 'r-', measured_data3, 'k-', measured_data4, 'g-')
        plt.show()
        # fig.canvas.draw()

############################################################################
# 3rd Level: Classes and Functions to Use SD1/M3xxxA and M5xxxA Instruments
############################################################################

class Sd1ModuleDescriptor:
    """
    Descriptor for SD1/M3xxxA module objects
    """
    def __init__(self, model_number, chassis_number, slot_number, options, engine_name):
        self.model_number = model_number
        self.chassis_number = chassis_number
        self.slot_number = slot_number
        self.options = options
        self.engine_name = engine_name


class DioModuleDescriptor:
    """
    Descriptor for M5xxxA module objects
    """
    def __init__(self, resource_id, query, reset, options, model_number, engine_name):
        self.resource_id = resource_id
        self.query = False
        self.reset = False
        self.options = options
        self.model_number = model_number
        self.engine_name = engine_name


class SystemSyncModuleDescriptor:
    """
    Descriptor for System Sync Modules
    """
    def __init__(self, resource_id, options):
        self.resource_id = resource_id
        self.options = options


class SystemSyncModuleConnection:
    """
    Describes the connection between two SyncModule objects ssm1 and ssm2. SSMs have always one upstream port:
    ssm1.connectivity.systemsync_downstream[ssm1_downstream_connector_number].connection = ssm2.connectivity.systemsync_upstream[0]
    """
    def __init__(self, ssm1_chassis, ssm1_downstream_connector_number, ssm2_chassis):
        self.ssm1_chassis = ssm1_chassis
        self.ssm1_downstream_connector_number = ssm1_downstream_connector_number
        self.ssm2_chassis = ssm2_chassis


class Module:
    """
    Class defining a modular instrument object and its properties
    """
    def __init__(self, instrument_object, hvi_interface, num_channels):
        self.instrument = instrument_object
        self.hvi = hvi_interface
        self.num_channels = num_channels


def open_awg_modules(config):
    """ 
    Opens and creates each M3xxxA AWG instrument object.
    Returns a dictionary of module objects whose keys are the HVI engine names.
    Please check SD1 3.x User Guide for options to open the instrument objects.
    """
    # Initialize output variables
    num_modules = 0
    module_dict = {} # dictionary of modules

    # Create PXI instruments objects  
    for descriptor in config.awg_module_descriptors:
        instr_obj = keysightSD1.SD_AOU()
        instr_obj_options = descriptor.options + ',simulate=true' if config.hardware_simulated else descriptor.options
        id = instr_obj.openWithOptions(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, instr_obj_options)
        if id < 0:
            raise Exception("Error opening instrument in chassis: {}, slot: {}! Error code: {}, {}. Exiting...".format(descriptor.chassis_number, descriptor.slot_number, id, keysightSD1.SD_Error.getErrorMessage(id)))
        nCh = instr_obj.getOptions("channels")
        if nCh == "CH2":
            num_channels = 2
        elif nCh == "CH4":
            num_channels = 4
        else:
            raise Exception("PXI module in chassis {}, slot {} returned number of channels = {} which is incorrect. Exiting... ".format(instr_obj.getChassis(), instr_obj.getSlot(), nCh))
        module_dict[descriptor.engine_name] = Module(instr_obj, instr_obj.hvi, num_channels)
        num_modules += 1
    
    return module_dict


def open_dig_modules(config):
    """ 
    Opens and creates each M3xxxA Digitizer instrument object.
    Returns a dictionary of module objects whose keys are the HVI engine names.
    Please check SD1 3.x User Guide for options to open the instrument objects.
    """
    # Initialize output variables
    num_modules = 0
    module_dict = {} # dictionary of modules

    # Create PXI instruments objects  
    for descriptor in config.dig_module_descriptors:
        instr_obj = keysightSD1.SD_AIN()
        instr_obj_options = descriptor.options + ',simulate=true' if config.hardware_simulated else descriptor.options
        id = instr_obj.openWithOptions(descriptor.model_number, descriptor.chassis_number, descriptor.slot_number, instr_obj_options)
        if id < 0:
            raise Exception("Error opening instrument in chassis: {}, slot: {}! Error code: {}, {}. Exiting...".format(descriptor.chassis_number, descriptor.slot_number, id, keysightSD1.SD_Error.getErrorMessage(id)))
        # Get number of channels
        nCh = instr_obj.getOptions("channels")
        if nCh == "CH2":
            num_channels = 2
        elif nCh == "CH4":
            num_channels = 4
        else:
            raise Exception("PXI module in chassis {}, slot {} returned number of channels = {} which is incorrect. Exiting... ".format(instr_obj.getChassis(), instr_obj.getSlot(), nCh))
        # Populate module dictionary
        module_dict[descriptor.engine_name] = Module(instr_obj, instr_obj.hvi, num_channels)
        num_modules += 1
    
    return module_dict


def open_dio_modules(config):
    """ 
    Opens and creates all the necessary M5302A digital I/O instrument objects.
    Returns a dictionary of module objects whose keys are the HVI engine names.
    Please check M5302A User Guide for options to open the instrument objects.
    """
    # Initialize output variables
    num_modules = 0
    module_dict = {} # dictionary of modules

    # Create M5302A digital I/O instrument objects  
    for descriptor in config.dio_module_descriptors:
        instr_obj_options = 'Simulate=1' if config.hardware_simulated else descriptor.options
        instr_obj = denim.KtM5302x(descriptor.resource_id, descriptor.query, descriptor.reset, instr_obj_options)
        # Create HVI interface
        hvi = denim.KtM5302xHvi(instr_obj.get_instance())
        # DIO instruments have always 8 channels
        num_channels = 8
        module_dict[descriptor.engine_name] = Module(instr_obj, hvi, num_channels)
        num_modules += 1
    
    return module_dict


def configure_awg(config, awg_module, num_channels = 2):
    """ 
    awg_queue_waveform(awg_module, num_channels) queues a Gaussian waveform to all the num_channels channels of
    the AWG object awg_module that is passed to the function
    """  
    # Load waveform to AWG memory
    awg_module.waveformFlush() #memory flush
    wave = keysightSD1.SD_Wave()
    wfmNum = 0
    wave.newFromFile("Waveforms/Gaussian.csv")
    awg_module.waveformLoad(wave, wfmNum)

    for channel in range(1, num_channels+1):
        # AWG queue flush 
        awg_module.AWGstop(channel)
        awg_module.AWGflush(channel)
        # Set AWG mode
        awg_module.channelWaveShape(channel, keysightSD1.SD_Waveshapes.AOU_AWG)
        awg_module.channelAmplitude(channel, config.amplitude)
        # AWG configuration
        awg_module.AWGqueueConfig(channel, config.queue_mode)
        awg_module.AWGqueueSyncMode(channel, config.sync_mode)
        # Queue waveform to the selected channel
        awg_module.AWGqueueWaveform(channel, wfmNum, config.awg_trigger_mode, config.awg_start_delay, config.awg_num_cycles, config.awg_prescaler)
        awg_module.AWGstart(channel) # AWG starts and waits for an AWG trigger


def configure_digitizer(config, digitizer_module, num_channels = 4):
    """ 
    This method configures multiple DAQ channels for cyclic acquisitions
    """
    # Input settings
    prescaler = config.dig_prescaler 
    fullscale = config.fullscale
    num_cycles = config.dig_num_cycles
    points_per_cycle = config.acquisition_points_per_cycle
    acquisition_delay = config.acquisition_delay
    trigger_mode = config.dig_trigger_mode

    # DAQstop: stops any previous acquisition
    digitizer_module.DAQstopMultiple(config.mask)
    
    # Configure DAQ channels
    for channel in range (1, num_channels+1):
        # Resets the DAQ
        digitizer_module.DAQflush(channel)
        # Digitizer channel configuration
        digitizer_module.channelInputConfig(channel, fullscale, keysightSD1.AIN_Impedance.AIN_IMPEDANCE_50, keysightSD1.AIN_Coupling.AIN_COUPLING_DC)
        digitizer_module.channelPrescalerConfig(channel, prescaler)
        # DAQ acquisitions configuration
        digitizer_module.DAQconfig(channel, points_per_cycle, num_cycles, acquisition_delay, trigger_mode)
    
    return points_per_cycle


if __name__ == '__main__':
    main()